@font-face 
  font-family: 'Inconsolata'
  font-style: normal
  font-weight: 400
  src: local('Inconsolata'), local('Inconsolata-Regular'), url(../fonts/Inconsolata-Regular.ttf) format('truetype')

@font-face
  font-family: 'Inconsolata'
  font-style: normal
  font-weight: 700
  src: local('Inconsolata Bold'), local('Inconsolata-Bold'), url(../fonts/Inconsolata-Bold.ttf) format('truetype')

@font-face
  font-family: 'Lato'
  font-style: normal
  font-weight: 400
  src: local('Lato Regular'), local('Lato-Regular'), url(../fonts/Lato-Regular.ttf) format('truetype')

@font-face
  font-family: 'Lato'
  font-style: normal
  font-weight: 700
  src: local('Lato Bold'), local('Lato-Bold'), url(../fonts/Lato-Bold.ttf) format('truetype')

@font-face
  font-family: 'Roboto Slab'
  font-style: normal
  font-weight: 400
  src: local('Roboto Slab Regular'), local('RobotoSlab-Regular'), url(../fonts/RobotoSlab-Regular.ttf) format('truetype')

@font-face 
  font-family: 'Roboto Slab'
  font-style: normal
  font-weight: 700
  src: local('Roboto Slab Bold'), local('RobotoSlab-Bold'), url(../fonts/RobotoSlab-Bold.ttf) format('truetype')

