// ------------------------------------------------------------
// CONTRIBUTORS, PLEASE READ THIS!
// ------------------------------------------------------------
// This generates the RTD sticky badge for non RTD themes. As
// always, only files labeled "theme_*.sass should be edited".
// ------------------------------------------------------------
$border-box-sizing: false !default

@import wyrm_core/wy_variables
@import theme_variables
@import bourbon
@import neat
@import wyrm_core/mixin
@import wyrm_core/grid_settings
@import _theme_badge_fa
@import _theme_badge
