// In here are varibles used for sphinx_rtd_theme, they either add to or overwrite the default ones
// that are set in wyrm_core/wy_variables.sass. You'll find wyrm in bower_components if you're looking
// for a reference.

$font-awesome-dir:                    "../font/"
$static-img:                          "../img/"
$mathjax-color:                       $text-color

$headerlink-color:                    $text-color

// Code colors
$text-viewcode-color:                 $green
$text-codexref-color:                 $text-color

// Definition list colors
$class-color:                         $blue
$method-color:                        $gray

// Footer colors
$footer-color:                        $gray-light

// Menu colors
$menu-vertical-background-color:      $section-background-color

// Menu text colors
$menu-color:                          $gray
$menu-dark:                           lighten($menu-color,15%) !default
$menu-medium:                         lighten($menu-color,25%) !default
$menu-light:                          lighten($menu-color,45%) !default
$menu-lighter:                        lighten($menu-color,60%) !default

// Menu link colors
$menu-link-color:                     $text-color
$menu-link-dark:                      $text-dark
$menu-link-medium:                    $text-medium
$menu-link-light:                     $text-light
$menu-link-active:                    $white

// Navigation colors
$nav-background-color:                $menu-background-color
$nav-search-background-color:         $blue
$nav-search-color:                    $section-background-color
$nav-link-color:                      $blue
$nav-link-color-visited:              $purple
$nav-link-color-hover:                lighten($nav-link-color, 6%) !default
$nav-link-color-alt:                  hsl(33, 100%, 51%)

// Sidebar colors
$sidebar-background-color:            $table-stripe-color
$sidebar-border-color:                $table-border-color
$sidebar-title-background-color:      $table-border-color

// Sphinx highlight color
$highlight-color:                     $yellow

$base-font-family:                    "Lato", "proxima-nova", "Helvetica Neue", Arial, sans-serif
$custom-font-family:                  "Roboto Slab", "ff-tisa-web-pro", "Georgia", Arial, sans-serif
$custom-font-family2:                 Georgia, serif
$code-font-family:                    Consolas, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", Monaco, "Courier New", Courier, monospace
